// =============================================================================
// Electronic compass for Treo650 communicator
// (C) MAY 2008 Hynek Sladky
// =============================================================================

/**
ToDo:
- videl bych to na prikaz pro trvale posilani a prikaz pro jednorazove mereni
  napr. 
  +AZI=1 (jedno mereni)
  +AZI=2 (vice mereni)
  +AZI=6 (vice mereni s ladicimi informacemi)
- POZOR! vypada to, ze seriovy port je napajeny, i kdyz je Palm vypnuty!!!
- pro minimalizaci odberu kdyz se nemeri bude potreba udelat podporu pomaleho taktu nebo sleepu
  je potreba toto domyslet a prijit si i ruznou podporu serioveho portu...
  treba by jen stacilo dat preruseni na hranu a uspat...
  ale po probuzeni pak neni jiste, jaky hodinovy kmitocet bude...
  - odber bez mereni: 4.2mA
  - odber s merenim: 8.5mA
  - treba jen zkusit nejaky melky sleep, kdy bezi seriovy port
    zkusit vypnout ADC
- pri kalibraci zajistit, ze se otoci o celych 360 stupnu!!!
  - nejaky pokus jsem udelal, ale neni to uplne ono...
    je potreba to predelat a vyzkouset
    udelat nejaky algoritmus, ktery bude detekovat dotyk maxima...


Hotovo:
o uprintf netiskne hodnotu 0!!!
  asi budu muset udelat vlastni funkci pro tisk celeho cisla...
o preruseni pro UART (RX i TX)
o preruseni pro CCR0 - trvale samplovani
o doplnit vstup UARTu pro ovladani - jen kdyz se posle prikaz, tak se zapne kompas
o posilani pres preruseni (takhle se zbytecne ceka moc dlouho, nez se odeslou data!)
o obecne by asi bylo lepsi, kdyby se cetnost mereni ridila casovacem a prerusenim...
  pak pujdou udelat filtry na 50Hz apod. takto ve smycce je to cele dost nahodne (nejasne definovany okamzik samplovani)
o getADC (vubec pouziti ADC) v preruseni a v hlavnim programu neni mozny!
  bude potreba najit cestu, jak vytvorit dalsi preruseni pro samplovani napeti
  - napr. slo by, aby se v kroku 2 a 4 provedlo mereni vsech potrebnych kanalu
    protoze to zarizuje HW, tak neni potreba dalsi preruseni, pak budu vedet, ze dane hodnoty jsou v pameti k dispozici...
o pokud bude nutne, tak je mozne kalibrovat presneji: otoci se kolem osy o 360 stupnu a zjisti se velikost X a Y
  - z toho pak bude mozne normovat X a Y tak, aby mely stejnou velikost os (elipsa -> kruznice)
  - X slozka muze mit hodnoty -12..29, Y -12..20 -> asi bude nutne toto nejak zpracovat a zkalibrovat...
    mozna tam je nejaky ss offset, ktery by se dal vykompenzovat prictenim nejake hodnoty...
    X: +15 Y: +15
  - viz http://www.robot-electronics.co.uk/htm/cmps_cal.shtml
    kompas podle Hanzala se kalibruje ulozenim 4 hodnot (NESW) nebo pomalym otacenim o 360 stupnu
    - pomale otaceni je jednoduche, ale nemusi byt dostacujici:
      - nevim, jestli minima/maxima opravdu odpovidaji svetovym stranam
    - 4 svetove strany jsou dobre, ale slozite
    takze je mozne, ze cilem kalibrace je z sisoidu udelat kruznici...
  o pred rozvahou o kalibraci provest mereni bez Palma a s Palmem pro celou kruznici (po 15st.)
o asi je malo RAM a prepisuje se mi promenna setup!!!
  zrusim aspon prumerovaci buffer a budu prumerovat (nacitat) primo do promennych
o do DBG informaci pridat i teplotu (rozsirit pocet vzorkovanych kanalu)
o nastaveni kalibrace
  +CAL=<calX>,<calY>
o asi by bylo dobre, kdyby zakladni konfigurace byla ulozena v EEPROM/Flash a bylo 
  mozne ji zmenit/ulozit prikazem
  - napr. vsechny zmeny jen v RAM, ale po prikazu k ulozeni budou platne i po restartu
  +SAVE
  ulozeni do Flash
  +RST
  nastaveni defaultnich hodnot
**/

#include <msp430x12x2.h>

#include <stdio.h>
#include <stdarg.h>
#include <__vfprintf.h>

//#include <stdlib.h>

#include "macros.h"

// ====================================
// definice HW
// ------------------------------------
// PORT 1
#define TMP_3   __bit (P1OUT, __bit0)
#define TMP_2   __bit (P1OUT, __bit1)
#define PWR1l   __bit (P1OUT, __bit2)
#define PWR1    __bit (P1DIR, __bit2)
#define PWR2l   __bit (P1OUT, __bit3)
#define PWR2    __bit (P1DIR, __bit3)
//#define TCK
//#define TMS
//#define TDI
//#define TDO
// PORT 2
#define Vcap    Ain0
#define OUT2P   __bit (P2OUT, __bit1)
#define OUT2N   __bit (P2OUT, __bit2)
#define Vsig1   Ain3
#define LED     __bit (P2OUT, __bit4)
//#define Rosc
// PORT 3
#define OUT1P   __bit (P3OUT, __bit0)
#define OUT1N   __bit (P3OUT, __bit1)
#define OUT12P  __bit (P3OUT, __bit2)
#define OUT12N  __bit (P3OUT, __bit3)
//#define TxD
//#define RxD
//#define PWRcap  __bit (P3OUT, __bit6)
#define PWRcapl __bit (P3OUT, __bit6)
#define PWRcap  __bit (P3DIR, __bit6)
#define Vsig2   Ain7
// ------------------------------------
// kontrolni promenna
// ------------------------------------
volatile unsigned int CHECK;
// ====================================
// nastaveni programu
// ------------------------------------
typedef struct {
  int Xofs, Xamp; // rozsah hodnot X
  int Yofs, Yamp; // rozsah hodnot Y
  unsigned int mtim; // rychlost opakovani mereni 1/4us
  unsigned int mcnt; // pocet mereni mezi zobrazenim vysledku
  unsigned int cmd_to; // delka timeoutu pro prikazovou radku
} setupType;
const setupType def_setup = {
  0, 60,
  0, 60,
  13333, // 50 mereni/s pri 6 mericich krocich
  25,
  4
};
setupType setup;
// ====================================
// definice ASM funkci
// ------------------------------------
void PulseN (void);
void PulseP (void);
void ChargeCapPulse (unsigned int count);
// ====================================
// cteni/prevod ADC
// ------------------------------------
/**
unsigned short getADC (unsigned short channel) {
  ADC10CTL1 &= 0x0FFF;
  ADC10CTL1 |= channel<<12;
  ADC10CTL0 |= ENC | ADC10SC;
  while (ADC10CTL1 & ADC10BUSY);
  ADC10CTL0 &= 0xFFFC;
  return ADC10MEM;
}
**/
// ====================================
// preruseni pro CC0
// ------------------------------------
volatile /*static*/ char step;
/*static*/ char run;
volatile long int xa, ya;
volatile unsigned int xycnt;
volatile /*static*/ unsigned tim;
enum {Vcc2, Vtemp, Verefm, Verefp, Ain7, Ain6, Ain5, Ain4, Ain3, Ain2, Ain1, Ain0, AD_CNT};
volatile /*static*/ int adc[AD_CNT];
// ------------------------------------
void timerA0 (void) __interrupt[TIMERA0_VECTOR] {
static int adc0, adc1;
  // priprava na dalsi preruseni
  //TACCR0 += 10000; // 100Hz pri 4 taktech
  //TACCR0 += 13333; // 50Hz pri 6 taktech
  TACCR0 += setup.mtim;
  // kontrola mezi promennych
  if (step>5) {
    step=0;
    if (tim) tim--;
  }
  // vypnuti nabijeni
  PWRcap = 0;
  //LED = 0;
  if (!step || step==3) {
    // mereni analogovych vstupu
    ADC10DTC0 = 0;
    ADC10DTC1 = AD_CNT;
    ADC10SA = (unsigned int)adc;
    //ADC10CTL1 = INCH2 | INCH1 | INCH0 | ADC10DIV1 | ADC10DIV0 | ADC10SSEL1 | CONSEQ_1;
    ADC10CTL1 = INCH3 | INCH1 | INCH0 | ADC10DIV1 | ADC10DIV0 | ADC10SSEL1 | CONSEQ_1;
    ADC10CTL0 = ENC | ADC10SC | ADC10SHT0 | ADC10SR | MSC | ADC10ON;
  }
  if (run) {
    //LED = 1;
    // zpracovani jednotlivych kroku
    switch (step) {
      case 1:
        OUT1N = 1;
        PulseP ();
        OUT1N = 0;
        break;
      case 2:
        OUT2N = 1;
        PulseP ();
        OUT2N = 0;
        // ulozeni vysledku mereni
        adc0 = adc[Vsig1];
        adc1 = adc[Vsig2];
        break;
      case 4:
        OUT1P = 0;
        PulseN ();
        OUT1P = 1;
        break;
      case 5:
        OUT2P = 0;
        PulseN ();
        OUT2P = 1;
        // zpracovani vysledku mereni
        xa += adc[Vsig1] - adc0;
        ya += adc[Vsig2] - adc1;
        xycnt++;
        break;
    }
    if (step && step!=3) {
      // zapnuti nabijeni
      PWRcap = 1;
      //LED = 1;
    }
    //LED = 0;
  }
  // dalsi krok
  step++;
}
// ====================================
// seriova komunikace
// ====================================
int uputc (int c);
void uputs (char *s);
// ====================================
// preruseni pro prijem UART
// ------------------------------------
#define rxbuflen  24
/*static*/ char rxbuf[rxbuflen];
/*static*/ volatile short rxw, rxr;
/*static*/ volatile cmdlen;
/*static*/ volatile char cmd_to;
void UART_rx (void) __interrupt[USART0RX_VECTOR] {
char c;
  if (rxw>=rxbuflen) rxw=0;
  c=U0RXBUF;
  cmd_to=setup.cmd_to;
  if (c=='\n' || c=='\r') {
    if (rxw) {
      rxbuf[rxw]=0;
      uputc ('\r');
      uputc ('\n');
      cmdlen=rxw;
      rxw=0;
    }
  } else {
    if (c==':') c='='; // preklad kvuli klavesnici Palm Treo650
    rxbuf[rxw++]=c;
    uputc (c);
  }
}
// ====================================
// preruseni pro vysilani UART
// ------------------------------------
#define txbuflen  24
/*static*/ char txbuf[txbuflen];
/*static*/ volatile short txw, txr;
void UART_tx (void) __interrupt[USART0TX_VECTOR] {
  U0TXBUF=txbuf[txr++];
  if (txr>=txbuflen) txr=0;
  if (txw==txr)
    IE2 &= ~UTXIE0;
}
// ====================================
// odeslani znaku
// ------------------------------------
int uputc (int c) {
short nptr;
  nptr=txw+1; if (nptr>=txbuflen) nptr=0;
  while (nptr==txr) 
    IE2 |= UTXIE0;
  txbuf[txw++]=c;
  if (txw>=txbuflen) txw=0;
  IE2 |= UTXIE0;
}
/**
int uprintf (const char *fmt, ...) {
int n;
va_list ap;
__printf_t iod;
  va_start(ap, fmt);
  iod.is_string = 0;
  iod.maxchars = 80; //INT_MAX;
  iod.u.output_fn = uputc;
  n = __vfprintf(&iod, fmt, ap);
  va_end(ap);
  return n;
}
/**/
void uputs (char *s) {
  while (*s) {
    uputc (*s);
    s++;
  }
}
// ====================================
// vlastni funkce pro tisk cisel
// ------------------------------------
/**
void uputi (int i) {
char buf[5];
int pos;
  if (i<0) {uputc ('-'); i=-i;}
  pos=0;
  do {
    buf[pos++]=(i%10)+'0';
    i/=10;
  } while (i);
  while (pos) {
    pos--;
    uputc (buf[pos]);
  }
}
**/
// ====================================
// prevod cisla na text
// ------------------------------------
/**/
void ltoa (unsigned long val, char *ptr, unsigned char base) {
const char *hex="0123456789ABCDEF";
  do {
    *ptr++=hex[val%base];
    val/=base;
  } while (val);
  *ptr=0;
}
/**/
// ====================================
// prevod textu na cislo
// ------------------------------------
/**
long atol (char *s) {
long val;
char sign;
  val=0;
  sign='+';
  if (*s=='-' || *s=='+') {
    sign=*s;
    s++;
  }
  while (*s>='0' && *s<='9') {
    val*=10;
    val+=*s-'0';
    s++;
  }
  return val;
}
**/
int atoi (char *s) {
int val;
char sign;
  val=0;
  sign='+';
  if (*s=='-' || *s=='+') {
    sign=*s;
    s++;
  }
  while (*s>='0' && *s<='9') {
    val*=10;
    val+=*s-'0';
    s++;
  }
  return (sign=='-')?-val:val;
}
// ====================================
// formatovany vystup
// ------------------------------------
int uprintf (char *fmt, ...) {
// zakladni promenne
va_list arg;
char c;
char buf[12];
int cnt;
unsigned long val;
// parametry tisku
char width; // minimum field width
char fill; // fill character
char just; // left/right justification
char len; // long/short
char base; // number base
char sign; // znak pro znamekno/vlastni znamenko
// pomocne promenne
//int i;
char i;
char *ptr;

  va_start (arg , fmt);
  cnt=0;
  while (c=*fmt++) {
    if (c=='%') {
      // inicializace promennych
      width=0;
      fill=' ';
      just=0;
      sign=0;
      base=0;
      len=sizeof (int);
      // parsovani formatovaciho retezce
      c=*fmt++;
      if (!c) break;
      // nastaveni zarovnani, tisku znamenka
      if (c=='-') {
        just=1;
      } else
      if (c=='+') {
        sign='+';
      } else
      if (c==' ') {
        sign=' ';
      } else {
        fmt--;
      }
      c=*fmt++;
      // nastaveni sirky tisku
      if (c=='0') {
        fill='0';
        c=*fmt++;
      }
      while (c>='0' && c<='9') {
        width*=10;
        width+=c-'0';
        c=*fmt++;
      }
      // modifikatory
      if (c=='l') {
        len=sizeof (long);
        c=*fmt++;
      } else
      if (c=='h') {
        len=sizeof (short);
        c=*fmt++;
      }
      // vlastni konverzni znaky
      if (c=='c') {
        uputc (*(int*)arg);
        cnt++;
        arg+=sizeof (int);
      } else
      if (c=='s') {
        ptr = *(char**)arg;
        if (!just) {
          for (i=0; ptr[i]; i++);
          while (width>(char)i) {
            uputc (' ');
            cnt++;
            i++;
          }
        }
        while (*ptr) {
          uputc (*ptr++);
          cnt++;
        }
        arg+=sizeof (char*);
      } else
      if (c=='d' || c=='i' || c=='u') {
        base=10;
      } else
      if (c=='o') {
        base=8;
      } else
      if (c=='x' || c=='X') {
        base=16;
      } else {
        // sem spadne i znak '%'
        uputc (c);
        cnt++;
      }
      if (base) {
        val=0;
        val=(len==sizeof (int))?*(unsigned int*)arg:*(long*)arg;
        if (c=='d' || c=='i') {
          if (len==sizeof (int) && (val&0x8000)) ((int*)&val)[1]=0xFFFF;
          if (val&0x80000000) {
            sign='-';
            val=-(long)val;
          }            
        }
        ltoa (val, buf, base);
        for (i=0; buf[i]; i++);
        // misto na znamenko
        if (sign && width) width--;
        // vypln pri zarovnani vpravo
        if (!just && fill!='0') {
          while (width>i) {
            uputc (' ');
            cnt++;
            width--;
          }
        }
        // znamenko
        if (sign) {
          uputc (sign);
          cnt++;
        }
        // vypln '0'
        if (fill=='0') {
          while (width>i) {
            uputc ('0');
            cnt++;
            width--;
          }
        }
        // tisk vlastniho cisla
        while (i) {
          uputc (buf[--i]);
          cnt++;
          if (width) width--;
        }
        // vypln pri zarovnani vpravo
        while (width) {
          uputc (' ');
          cnt++;
        }
        arg+=len;
      }
    } else {
      uputc (c);
      cnt++;
    }
  }
  va_end (arg);
  return cnt;
}
// ====================================
// inicializace HW po zapnuti
// ------------------------------------
void HWinit (void) {
  // vypnuti WDT
  WDTCTL = WDTPW | WDTHOLD;
  // nastartovani LFXT1 (32kHz)
  //BCSCTL1 = RSEL2; // ACLK=LFXT, pasmo 1MHz
  //BCSCTL1 = RSEL2 | RSEL1 | RSEL0; // ACLK=LFXT, pasmo 4MHz
  BCSCTL1 = DIVA1 | RSEL2 | RSEL1 | RSEL0; // ACLK=LFXT/4, pasmo 4MHz
  // zakladni nastaveni DCO
  DCOCTL = DCO1 | DCO0 | MOD4;
  // nastaveni Rext, MCLK=DCOCLK, SMCLK=DCOCLK
  //BCSCTL2 = DCOR | DIVS1 | DIVS0; // external resistor, SMCLK=DCOCLK/8
  BCSCTL2 = DCOR; // external resistor
  //BCSCTL2 = 0; // internal resistor
  // nastaveni portu
  P1OUT = 0b00001100;
  P1DIR = 0b11110011;
  P1SEL = 0b00000000;
  P2OUT = 0b00000010;
  P2DIR = 0b11010110;
  P2SEL = 0b00101001;
  P3OUT = 0b01110101;
  P3DIR = 0b00111111;
  P3SEL = 0b10000000;
/**
  // cekani na nabeh oscilatoru
  do
    IFG1 &= ~OFIFG; // clear oscfault flag
  while ((IFG1 & OFIFG) != 0);
**/
  // nastaveni WDT casovace
  WDTCTL = WDTPW | WDTTMSEL | WDTSSEL | WDTIS0; // timer - aclk - 8192
  // nastaveni casovace A
  TACTL = TASSEL1 | MC1; // SMCLK, 0xFFFF
  //TACTL = TASSEL1 | MC1 | ID0; // SMCLK, 0xFFFF
  //TACTL = TASSEL0 | MC1; // ACLK, 0xFFFF
  // nastaveni casovace pro spousteni mereni
  TACCTL0 = CCIE;
  // nastaveni casovace 1 pro 1ms cekaci funkci
  TACCTL1 = 0;
  // nastaveni casovace 2 pro doladovani DCO
  TACCTL2 = CM0 | CCIS0 | SCS | CAP;
  // nastaveni UART0
  P3SEL |= 0x10; //0x30;
  // UART
  U0CTL = SWRST;
  U0CTL = CHAR | SWRST; // 8-bit
  U0TCTL = SSEL1; // SMCLK
  //baud = 4000000/baud;
  U0BR1 = ((4000000/19200)>>8)&0xFF;
  U0BR0 = (4000000/19200)&0xFF;
  //RS_MCTL = 0b01101011;//107;
  ME2 = UTXE0; // | RS_RXE;
  //IE2 = URXIE0;
  U0CTL &= ~SWRST;
  rxw=rxr=txw=txr=0;

  // nastaveni ADC
  //ADC10CTL0 = SREF0 | ADC10SHT1 | ADC10SHT0 | ADC10SR | REF2_5V | REFON | ADC10ON;
  //ADC10CTL0 = ADC10SHT1 | ADC10SHT0 | ADC10SR | REF2_5V | REFON | ADC10ON;
  ADC10CTL0 = ADC10SHT0 | ADC10SR | MSC | ADC10ON;
  //ADC10CTL1 = ADC10DIV1 | ADC10DIV0 | ADC10SSEL1;
  ADC10CTL1 = INCH2 | INCH1 | INCH0 | ADC10DIV1 | ADC10DIV0 | ADC10SSEL1 | CONSEQ_1;
  ADC10AE=0b10001001;
  ADC10DTC0 = 0;

  // povoleni preruseni
  //_BIS_SR (GIE);
}
// ====================================
// priprava HW na usnuti
// ------------------------------------
void HWsleep (void) {
  P1OUT = 0b00001100;
  P1DIR = 0b11110011;
  P1SEL = 0b00000000;
  P2OUT = 0b00000010;
  P2DIR = 0b11010110;
  P2SEL = 0b00101001;
  P3OUT = 0b01110101;
  P3DIR = 0b00111111;
  P3SEL = 0b10000000;
  ADC10CTL0 = 0;
  ADC10CTL1 = 0;
  TACCTL0 = 0;
  TACCTL1 = 0;
  TACCTL2 = 0;
  TACTL = 0;
  IE2 = 0;
  ME2 = 0;
}
// ====================================
// casovaci funkce
// ------------------------------------
void delay (unsigned int ms) {
/**
  TACCR1=TAR+(ms*3277L)/100;
  TACCTL1 &= ~CCIFG;
  while (!(TACCTL1 & CCIFG));
**/
  TACCR1=TAR;
  while (ms) {
    TACCR1+=4000;
    TACCTL1 &= ~CCIFG;
    ms--;
    while (!(TACCTL1 & CCIFG));
  }
}
// ====================================
// doladeni pozadovane frekvence
// ------------------------------------
void TuneDCO (unsigned short val) {
unsigned short tim;
  // doladovani frekvence
  do {
    // cekam na prvni hranu
    TACCTL2 &= ~CCIFG;
    while (!(TACCTL2 & CCIFG));
    // prectu capture
    tim = TACCR2;
    // cekam na dalsi hranu
    TACCTL2 &= ~CCIFG;
    while (!(TACCTL2 & CCIFG));
    // vypocitam rozdil
    tim = TACCR2 - tim;
    // podle vysledku doladim DCO
    // jeste by slo doplnit zmenu RSEL pro sirsi oblast...
    if (tim>val) DCOCTL--;
    else
    if (tim<val) DCOCTL++;
    else break;
  } while (1);
}
// ====================================
// nabiti kapacity
// ------------------------------------
void ChargeCap (void) {

  // dokud je napeti mensi nez polovina, nabijim pulzne
  while (adc[Vcap]<512) {
    ChargeCapPulse (1); // -> cca 125 pulzu pro polovinu napeti z 0V
  }
  // ted uz muzu nabijet trvale
  PWRcap = 1;
  // cekam do 95% napeti (20ms?)
  while (adc[Vcap]<980);
  // pro uplne nabiti by se muselo cekat 2s
  //while (getADC (Vcap)<1023) cnt++;
  PWRcap = 0;
}
// ====================================
// funkce pro zjednoduseny vypocet uhlu
// ------------------------------------
int arctg (int x, int y) {
int ofs, tmp, sign;

  // zjisteni kvadrantu a vypocet absolutnich hodnot
  ofs=0;
  sign=1;
  if (y<0) {y=-y; x=-x; ofs+=180;}
  if (x<0) {tmp=x; x=y; y=-tmp; ofs+=90;}
  if (y>x) {tmp=x; x=y; y=tmp; ofs+=90; sign=-sign;}
  // ted uz staci jen nanormovat 0-45
  tmp = (y * 45)/x;
  // a ted zpatky rozkodovat kvadrant
  tmp = (tmp*sign)+ofs;
  while (tmp<0) tmp+=360;
  while (tmp>=360) tmp-=360;
  return tmp;
}
// ====================================
// porovnani retezce
// ------------------------------------
int strcmp (char *s1, char *s2) {
  while (*s1 && *s1==*s2 && *s2 ) {
    s1++;
    s2++;
  }
  if (!*s1 && !*s2) return 0;
  return (*s1>*s2)?1:-1;
}
int strncmp (char *s1, char *s2, int len) {
  while (len && *s1 && *s1==*s2 && *s2 ) {
    s1++;
    s2++;
    len--;
  }
  if (!len || (!*s1 && !*s2)) return 0;
  return (*s1>*s2)?1:-1;
}
void memcpy (void *dst, const void *src, unsigned int len) {
unsigned char *d, *s;
  s=(unsigned char*)src;
  d=dst;
  while (len) {
    *d=*s;
    d++;
    s++;
    len--;
  }
}
char *NextNum (char *s) {
  if (*s=='-' || *s=='+') s++;
  while (*s>='0' && *s<='9') s++;
  while (*s && (*s<'0' || *s>'9') && *s!='-' && *s!='+') s++;
  return s;
}
// ====================================
// ulozeni obsahu do Flash
// ------------------------------------
int WriteFlash (void *dst, void *src, unsigned int len) {
unsigned char *s, *d;
unsigned int u;
  s=src;
  d=dst;
  u=len;
  while (u && *s==*d) {
    s++;
    d++;
    u--;
  }
  if (!u) return 0;
  _BIC_SR (GIE);
  // smazani Flash
  FCTL2 = FWKEY | FSSEL0 | 11;
  FCTL3 = FWKEY;
  FCTL1 = FWKEY | ERASE;
  //FCTL3 = FWKEY | LOCK;
  *(unsigned char*)dst=0; // dummy write
  FCTL1 = FWKEY;
  // naprogramovani Flash
  //FCTL3 = FWKEY;
  FCTL1 = FWKEY | WRT;
  s=src;
  d=dst;
  u=len;
  while (u) {
    *d=*s;
    d++;
    s++;
    u--;
  }
  FCTL1 = FWKEY;
  FCTL3 = FWKEY | LOCK;
  _BIS_SR (GIE);
  // verifikace Flash
  return 0;
}
// ====================================
// hlavni program
// ------------------------------------
void main (void) {
// promenne pro zpracovani signalu
int adc0, adc1;
unsigned short cnt;
char *s;
long avgx, avgy;
unsigned char cmask, ccnt;

  // po pripojeni k Palmovi se ceka min 150ms pred pripojenim napajeni
  // pak se po dobu min 50ms ctou signaly RxD, TxD (musi byt L)
  // pripoji se napajeni a po 200ms se po dobu 50ms ctou signaly RxD, TxD (musi byt H)
  // tj. po zapnuti se musi cekat aspon 250ms, nez se zacne komunikovat...

  CHECK=0xDEAD;
  // inicializace HW
  HWinit ();
  // zakladni nastaveni DCO podle krystalu
  TuneDCO (488);
  // cekani 250ms (pro jistotu, abych nemusel merit cas od resetu)
  delay (250);
  // nastaveni promennych
  cmdlen=0;
  run=0;
  //memcpy (&setup, &def_setup, sizeof (setup));
  memcpy (&setup, (void*)0x1000, sizeof (setup));
  // povoleni prijmu serioveho portu
  P3SEL |= 0x20;
  ME2 |= URXE0;
  IE2 |= URXIE0;
  // povoleni preruseni
  _BIS_SR (GIE);

  // docasny prikaz po zapnuti
  //rxbuf[0]='r'; rxbuf[1]='u'; rxbuf[2]='n'; rxbuf[3]=0; cmdlen=3;
  // hlavni smycka programu = kooperativni multitasking
  do {
    //prijem prikazu
    if (cmdlen) {
      if (!strncmp (rxbuf, "+AZI=", 5)) {
        if (rxbuf[5]=='?') {
          uprintf ("*AZI=%d\r\n", run);
        } else {
          // nacteni parametru
          adc0 = atoi (&rxbuf[5]);
          if (adc0) {
            // zapnuti HW
            PWR1 = 1;
            delay (50);
            PWR2 = 1;
            delay (50);
            // nulovani promennych pro prumer
            xa=ya=0;
            xycnt=0;
            // zakladni nabiti kapacity
            ChargeCap ();
            PWRcap = 1;
            tim=setup.mcnt;
            // povoleni prevodu
            run=adc0;
          } else {
            // zakazani prevodu
            run=0;
            // vypnuti HW
            PWRcap = 0;
            PWR1 = 0;
            PWR2 = 0;
          }
          //uprintf ("*AZI=%d\r\n", run);
        }
      } else
      if (!strcmp (rxbuf, "+INFO")) {
        // informace o programu
        uprintf ("*INFO:Compass v0.1\r\n");
      } else
      if (!strncmp (rxbuf, "+CNT=", 5)) {
        if (rxbuf[5]=='?') {
          uprintf ("*CNT=%d\r\n", setup.mcnt);
        } else {
          setup.mcnt = atoi (&rxbuf[5]);
          //uprintf ("*CNT=%d\r\n", setup.mcnt);
        }
      } else
      if (!strncmp (rxbuf, "+TIM=", 5)) {
        if (rxbuf[5]=='?') {
          uprintf ("*TIM=%d\r\n", setup.mtim);
        } else {
          setup.mtim = atoi (&rxbuf[5]);
          //uprintf ("*TIM=%d\r\n", setup.mtim);
        }
      } else
      if (!strncmp (rxbuf, "+CAL", 4) && (rxbuf[4]=='=' || !rxbuf[4])) {
        if (rxbuf[4]=='=') {
          if (rxbuf[5]=='?') {
            // dotaz na kalibraci
            uprintf ("*CAL=%d,%d,%d,%d\r\n", setup.Xofs, setup.Xamp, setup.Yofs, setup.Yamp);
          } else {
            // rucni nastaveni kalibrace
            s=&rxbuf[5];
            setup.Xofs=atoi (s);
            if (*(s=NextNum (s)))
              setup.Xamp=atoi (s);
            if (*(s=NextNum (s)))
              setup.Yofs=atoi (s);
            if (*(s=NextNum (s)))
              setup.Yamp=atoi (s);
          }
        } else {
          // automaticka kalibrace
          if (run) {
            uprintf ("*CAL:turn slowly 360 degree\r\n");
            setup.Xamp=setup.Yamp=-2000; setup.Xofs=setup.Yofs=2000;
            run|=0x40;
            cmask=ccnt=0;
          } else {
            uprintf ("*ERROR:compass in idle mode!\r\n");
          }
        }
      } else
      if (!strcmp (rxbuf, "+SAVE")) {
        // ulozeni hodnot do Flash
        WriteFlash ((void*)0x1000, &setup, sizeof (setup));
        uprintf ("*SAVE:done.\r\n");
      } else
      if (!strcmp (rxbuf, "+INIT")) {
        // nastaveni defaultnich hodnot
        memcpy (&setup, &def_setup, sizeof (setup));
        uprintf ("*INIT:done.\r\n");
      } else
      if (!strcmp (rxbuf, "+RST")) {
        // pokus o reset
        WDTCTL = 0x1234;
      } else {
        uprintf ("*ERROR: Unknown command!\r\n");
      }
      cmdlen=0;
    }
    if (!tim) {
      tim=setup.mcnt;
      if (run && !rxw) {
        LED = 1;
        // vypocet a poslani vysledku
        _BIC_SR (GIE);
        cnt=xycnt;
        avgx=xa;
        avgy=ya;
        xa=ya=0;
        xycnt=0;
        _BIS_SR (GIE);
        adc0+=cnt/2;
        adc1+=cnt/2;
        adc0=avgx/cnt;
        adc1=avgy/cnt;
        if (run&0x40) {
          if (adc0>setup.Xamp) setup.Xamp=adc0;
          if (adc0<setup.Xofs) setup.Xofs=adc0;
          if (adc1>setup.Yamp) setup.Yamp=adc1;
          if (adc1<setup.Yofs) setup.Yofs=adc1;
          //if (adc0+10<setup.Xamp && adc0-10>setup.Xofs && adc1+10<setup.Yamp && adc1-10>setup.Yofs) {
          if (adc0+10>setup.Xamp && (cmask&0x03)==0x00) cmask|=0x01;
          if (adc0+10<setup.Xamp && (cmask&0x03)==0x01) {ccnt++; cmask|=0x02;}
          if (adc0-10<setup.Xofs && (cmask&0x0C)==0x00) cmask|=0x04;
          if (adc0-10>setup.Xofs && (cmask&0x0C)==0x04) {ccnt++; cmask|=0x08;}
          if (adc1+10>setup.Yamp && (cmask&0x30)==0x00) cmask|=0x10;
          if (adc1+10<setup.Yamp && (cmask&0x30)==0x10) {ccnt++; cmask|=0x20;}
          if (adc1-10<setup.Yofs && (cmask&0xC0)==0x00) cmask|=0x40;
          if (adc1-10>setup.Yofs && (cmask&0xC0)==0x40) {ccnt++; cmask|=0x80;}
          if (cmask==0xFF) cmask=0;
          //uprintf ("*CAL:%d\r\n", cmask);
          //if (cmask==0x33) {
          uprintf ("*CAL:%d\r\n", ccnt);
          if (ccnt>=6) {
            setup.Xofs = (setup.Xamp + setup.Xofs)/2;
            setup.Xamp -= setup.Xofs;
            setup.Yofs = (setup.Yamp + setup.Yofs)/2;
            setup.Yamp -= setup.Yofs;
            uprintf ("*CAL:done.\r\n");
            run&=~0x40;
          }
        } else {
          adc0-=setup.Xofs;
          adc1-=setup.Yofs;
          // poslani vysledku
          uprintf ("*AZI:%d\r\n", arctg (adc0, (adc1*setup.Xamp)/setup.Yamp));
        }
        if (run>1) {
          //uprintf ("*DBG:%d,%d,%d,%d\r\n", adc0, adc1, adc[Vtemp],setup.mcnt);
          uprintf ("*DBG:%d,%d,%d\r\n", adc0, adc1, adc[Vtemp]);
        }
        LED = 0;
      }
    }
    if (IFG1&WDTIFG) {
      IFG1 &= ~WDTIFG;
      if (!run) {
        // pro ladeni blikam LEDkou
        LED = 1;
        delay (1);
        LED = 0;
      }
      // kazdou 1s doladim DCO podle krystalu
      TuneDCO (488);
      // pokud dlouho visi prikazova radka, smazu ji
      if (cmd_to) {
        cmd_to--;
        if (!cmd_to && rxw) {
          rxw=0;
          uputs ("\r\n");
        }
      }
    }
    if (CHECK!=0xDEAD) {
      run=0;
      uprintf ("*ERROR:Stack overflow!\r\n");
    }
  } while (1);
  // usnuti - asi neni funkcni, protoze jsem ho neladil...
  HWsleep ();
  _BIS_SR(LPM3_bits);
}
